import numpy as np
import uncertainties
from uncertainties import ufloat
from uncertainties.umath import *

from copy import *

UF = "<class 'uncertainties.core.Variable'>"
UFAS = "<class 'uncertainties.core.AffineScalarFunc'>"
UFlist = [UF, UFAS]

# 아래의 함수들은 fitting을 위한 함수들이다.


def Coefficient_of_Determination_linear(a, b, X, Y):  # 불확도 있는 리스트다.
    """
    The function return the value of the Coefficient of Determination of linear fitting.

    Parameter
        a : number, ufloat, The slope value of fitting function.
        b : number, ufloat, The intercept of fitting function.
        X : list of number or ufloat, The x-axis data.
        Y : list of number or ufloat, The y-axis data.
    """

    # Inputs type Error Handling
    if isinstance(a, (int, float)) or str(type(a)) in UFlist:  # check 'a' is calculatable
        pass
    else:
        raise TypeError("Input a should be a number or ufloat.")
    if isinstance(b, (int, float)) or str(type(b)) in UFlist:  # check 'b' is calculatable
        pass
    else:
        raise TypeError("Input b should be a number or ufloat.")
    try:  # check 'X' is attributable
        X[0]
    except TypeError:
        raise TypeError("Input X should be a list of numbers or a list of ufloats.")
    try:  # check 'Y' is attributable
        Y[0]
    except TypeError:
        raise TypeError("Input Y should be a list of numbers or a list of ufloats.")
    N = len(X)
    if N == len(Y):  # check the length of 'X' and 'Y'
        pass
    else:
        raise ValueError("Input X and Y should have the same length.")
    XX = copy(X)
    YY = copy(Y)
    i = 0
    while i < N:
        if isinstance(X[i], (int, float)):  # check the elements of 'X' is calculatable
            pass
        elif str(type(X[i])) in UFlist:
            XX[i] = X[i].n
        else:
            raise TypeError("Input X should have ONLY number elements or ufloat elements.")
        if isinstance(Y[i], (int, float)):  # check the elements of 'Y' is calculatable
            pass
        elif str(type(Y[i])) in UFlist:
            YY[i] = Y[i].n
        else:
            raise TypeError("Input Y should have ONLY number elements or ufloat elements.")
        i += 1
    
    y_i = [a * XX[i] + b for i in range(N)]
    SS_tot = np.sum([(YY[i] - (np.sum(YY) / N))**2 for i in range(N)])
    SS_res = np.sum([(y_i[i] - YY[i])**2 for i in range(N)])
    R2 = 1 - (SS_res / SS_tot)
    return R2


def linear_fitting(X, Y):  # scalarlist의 fitting
    """
    The function return the slope and intercept of fitting graph.
    This function consider only scalar value. Therefore, the uncertainties of data are ignored.

    Parameter
        X : list of numbers, the x-axis data.
        Y : list of numbers, the y-axis data.
    
    Return value
        [slope(ufloat), intercept(ufloat), Coefficient of Determination(number)]
    """
    
    try:  # check 'X' is attributable
        X[0]
    except TypeError:
        raise TypeError("Input X should be a list of numbers.")
    try:  # check 'Y' is attributable
        Y[0]
    except TypeError:
        raise TypeError("Input Y should be a list of numbers.")
    n = len(X)
    if n == len(Y):  # check the length of 'X' and 'Y'
        pass
    else:
        raise ValueError("Inputs 'X' and 'Y' must have same length.")
    # Inputs type Error Handling
    i = 0
    while i < n:
        # check the elements of 'X' and 'Y' are numbers.
        if isinstance(X[i], (int, float)) and isinstance(Y[i], (int, float)):
            pass
        else:
            ValueError("Inputs 'X' and 'Y' should be lists of numbers.")
            break
        i += 1
    
    Sx = np.sum(X)
    Sx2 = np.sum([X[i] * X[i] for i in range(n)])
    S = n
    Sy = np.sum(Y)
    Sxy = np.sum([X[i] * Y[i] for i in range(n)])
    a = (S * Sxy - Sx * Sy) / (Sx2 * S - Sx * Sx)  # slope
    b = (-Sx * Sxy + Sx2 * Sy) / (Sx2 * S - Sx * Sx)  # intercept
    SS = np.sqrt(sum([(Y[i] - a * X[i] - b)**2 for i in range(n)]) / (n - 2))
    da = SS * np.sqrt(n / (n * Sx2 - Sx * Sx))  # the S.E. of slope
    db = SS * np.sqrt(Sx2 / (n * Sx2 - Sx * Sx))  # the S.E. of intercept
    R2 = Coefficient_of_Determination_linear(a, b, X, Y)
    return [ufloat(a, da), ufloat(b, db), R2]


def linear_fitting_Yerr(X, Y, Yerr):  # Y만 ufloatlist인 경우의 fitting
    """
    The function return the slope and intercept of fitting graph.
    This function consider the case that ONLY y data have uncertainties.

    Parameter
        X : list of numbers, the x-axis data.
        Y : list of numbers, the y-axis data.
        Yerr : list of numbers, the uncertainties of y-axis data.
    
    Return value
        [slope(ufloat), intercept(ufloat), Coefficient of Determination(number)]
    """
    # Inputs type Error Handling
    try:
        X[0]
    except TypeError:
        raise TypeError("Input X should be a list of numbers.")
    try:
        Y[0]
    except TypeError:
        raise TypeError("Input Y should be a list of numbers.")
    try:
        Yerr[0]
    except TypeError:
        raise TypeError("Input Yerr should be a list of numbers.")
    n = len(X)
    if n == len(Y) and n == len(Yerr):
        pass
    else:
        raise ValueError("Inputs 'X' and 'Y', 'Yerr' should have same length.")
    i = 0
    while i < n:
        if isinstance(X[i], (int, float)) and isinstance(Y[i], (int, float)) and isinstance(Yerr[i], (int, float)):
            pass
        else:
            ValueError("Inputs 'X' and 'Y', 'Yerr' should be lists of numbers.")
            break
        i += 1
    Yerr2 = [x * x for x in Yerr]
    Sx = np.sum([X[i] / Yerr2[i] for i in range(n)])
    Sx2 = np.sum([X[i] * X[i] / Yerr2[i] for i in range(n)])
    S = np.sum([1 / Yerr2[i] for i in range(n)])
    Sy = np.sum([Y[i] / Yerr2[i] for i in range(n)])
    Sxy = np.sum([X[i] * Y[i] / Yerr2[i] for i in range(n)])
    a = (S * Sxy - Sx * Sy) / (Sx2 * S - Sx * Sx)  # 기울기
    b = (Sxy - a * Sx2) / Sx  # 절
    da = np.sqrt(S / (Sx2 * S - Sx * Sx))
    db = np.sqrt(Sx2 / (Sx2 * S - Sx * Sx))
    R2 = Coefficient_of_Determination_linear(a, b, X, Y)
    return [ufloat(a, da), ufloat(b, db), R2]


def York_func(X, Y, m):  # X, Y 모두 ufloatlist인 경우의 fitting 을 위한 함수
    """
    A function to use when iterate the slope with York's method fitting.

    Parameter
        X : A list of ufloats. x-data to fit.
        Y : A list of ufloats. y-dta to fit.
        m : int or float. The guess value of Slope.
    
    Return
        [slope, intercept, square of stdv of slope, square of stdv of intercept, SE of slope, SE of intercept, CoD(R^2)]
    """

    # York, D.(1966)
    # Residual(difference vector with data and polint on fitting line) is
    # perpendicular with fitting line

    # Weight Matrix (Quadratic Form) is the following.
    # [s^2(X_k)   s(X_k Y_k)]
    # [s(X_k Y_k)   s^2(Y_k)]

    # When the notation of Residual vector is V_k,
    # V_k^T W_k V_k is the SSE of each data
    # The Sum of SSE of each data is S (SSE of all data)(must be least)
    try:
        X[0]
    except TypeError:
        raise TypeError("Input 'X' should be a list of ufloats.")
    try:
        Y[0]
    except TypeError:
        raise TypeError("Input 'Y' should be a list of ufloats.")
    if len(X) != len(Y):
        raise TypeError("Input 'X' and 'Y' should have same length.")
    N = len(X)
    i = 0
    while i < N:
        if str(type(X[i])) not in UFlist:
            raise TypeError("The elements of 'X' should be ufloats : X[i]=", X[i])
        if str(type(Y[i])) not in UFlist:
            raise TypeError("The elements of 'Y' should be ufloats.")
        i += 1
    if isinstance(m, (int, float)):
        pass
    else:
        raise TypeError("Input 'm' should be calculatable.")
    
    n = len(X)
    LN = range(n)
    nX = [X[i].n for i in LN]  # data of X
    sX = [X[i].s for i in LN]  # Standard Error(S.E) of X
    nY = [Y[i].n for i in LN]  # data of Y
    sY = [Y[i].s for i in LN]  # Standard Error(S.E) of Y

    wX = [1 / (sX[i] * sX[i]) for i in LN]  # The weight of X datas
    wY = [1 / (sY[i] * sY[i]) for i in LN]  # The weight of Y datas

    W = [(wX[i] * wY[i]) / (wX[i] + m * m * wY[i]) for i in LN]
    # The Weight Function of each point.
    # It is different with Weight Matrixisinstance(X[i], (int, float)

    sW = np.sum(W)  # sum of W

    x_bar = np.sum([W[i] * nX[i] for i in LN]) / sW
    # Weighted Average of X
    y_bar = np.sum([W[i] * nY[i] for i in LN]) / sW
    # Weighted Average of Y

    U = [nX[i] - x_bar for i in LN]
    # Difference with data and weighted average
    V = [nY[i] - y_bar for i in LN]

    B = [W[i] * (U[i] / wY[i] + m * V[i] / wX[i]) for i in LN]
    # Beta in paper written by York,D.(1966)
    B_bar = np.sum([W[i] * B[i] for i in LN]) / sW
    # Weighted average of Beta

    M = np.sum([W[i] * B[i] * V[i] for i in LN]) / np.sum([W[i] * B[i] * U[i] for i in LN])
    # Iterated Slope
    a = y_bar - M * x_bar
    # The Weighted averages are satisfy 'y=Mx+a'

    res_x = [B[i] - U[i] for i in LN]  # residual of x datas
    res_y = [M * B[i] - V[i] for i in LN]  # residual of y datas

    D = np.sum([W[i] * U[i] * V[i] for i in LN]) / M + 4 * np.sum([W[i] * res_x[i] * (B[i] - B_bar) for i in LN])
    # D in paper written by York,D.(1966)
    s2_m = np.sum([U[i] * U[i] / wY[i] + V[i] * V[i] / wX[i] for i in LN]) / (D * D)
    s2_a = ((2 * B_bar + x_bar)**2) * s2_m + 2 * (2 * B_bar + x_bar) * B_bar / D + 1 / sW

    S = np.sum([wX[i] * res_x[i] * res_x[i] + wY[i] * res_y[i] * res_y[i] for i in LN])
    # Total SSE

    SE_m = np.sqrt(s2_m * S / (n - 2))  # Standard Error of Slope
    SE_a = np.sqrt(s2_a * S / (n - 2))  # Standard Error of intercept
    # np.sqrt(S/n-2) is 'goodness of fitting'
    R2 = Coefficient_of_Determination_linear(M, a, X, Y)
    # Determination Coefficient
    return [M, a, s2_m, s2_a, SE_m, SE_a, R2]


def York_deri(X, Y, m, dm):
    funcp = York_func(X, Y, m + dm)[0] - (m + dm)
    funcm = York_func(X, Y, m - dm)[0] - (m - dm)
    return (funcp - funcm) / (2 * dm)


def York_Newton_func(X, Y, m, dm):
    """
    A function to use in Newton Method to find Slope with York's method.
    """
    func = York_func(X, Y, m)[0] - m
    Ans = m - func / York_deri(X, Y, m, dm)
    return Ans


def linear_fitting_York(X, Y, dmratio=1e6, limit=100, initguess='normal', pr=False):
    """
    A function which find fitting function of two ufloat lists with York's method.

    Parmater
        X : A list of ufloats. x-data to fit.
        Y : A list of ufloats. y-data to fit.
        dmratio : The relative interval to calculate derivative of York_func.
        limit : The limit number of iteration of Newton Method.
        initguess : Method to determine Initial Guess value of Newton Method finding slope. 'normal' or 'yerr'
            'normal' : Initguess with fitting without considering error.
            'yerr'   : Initguess with fitting with considering only y-data error.
        pr : Whether print the Return value. True or False.
    
    Return
        [slope, intercept, square of stdv of slope, square of stdv of intercept, SE of slope, SE of intercept, CoD(R^2)]
    """
    
    # input type Error Handling
    try:
        X[0]
    except TypeError:
        raise TypeError("Input X should be a list of ufloats.")
    try:
        Y[0]
    except TypeError:
        raise TypeError("Input Y should be a list of ufloats.")
    N = len(X)
    if N == len(Y):
        pass
    else:
        raise ValueError("Input X and Y should have the same length.")
    UF = "<class 'uncertainties.core.Variable'>"
    i = 0
    while i < N:
        if str(type(X[i])) in UFlist:
            pass
        else:
            raise TypeError("Input X should have ONLY ufloat elements.")
        if str(type(Y[i])) in UFlist:
            pass
        else:
            raise TypeError("Input Y should have ONLY ufloat elements.")
        i += 1
    
    # initguess parameter Error Handling
    if initguess == 'normal':
        nX = [x.n for x in X]
        nY = [y.n for y in Y]
        m = linear_fitting(nX, nY)[0].n
    elif initguess == 'yerr':
        nX = [x.n for x in X]
        nY = [y.n for y in Y]
        Yerr = [y.s for y in Y]
        m = linear_fitting_Yerr(nX, nY, Yerr)[0].n
    else:
        raise ValueError("Parameter initguess should be 'normal' or 'yerr'.")
    
    # Newton Method Part to Find the Correct Slope.
    i = 0
    dm = m / dmratio
    M_0 = m
    M = [M_0]
    while True:
        M.append(York_Newton_func(X, Y, M[i], dm))
        if np.abs(York_deri(X, Y, M[i], dm)) <= 10**(-10):
            raise ZeroDivisionError("The derivative became 0")
            break
        if i > 1:
            if np.abs(M[i] - M[i - 1]) <= 10**(-9):
                Ans = York_func(X, Y, M[i])
                if pr is True:
                    print("The Solution is", M[i])
                    print("The number of iteration:", i)
                    print('The Slope is {:.2uP}'.format(ufloat(Ans[0], Ans[4])))
                    print('The Intercept is {:.2uP}'.format(ufloat(Ans[1], Ans[5])))
                    print('The Sigma^2 of Slope is', Ans[2])
                    print('The Sigma^2 of Intercept is', Ans[3])
                    print('The Value of R^2 if', Ans[6])
                return York_func(X, Y, M[i])
                break
            else:
                pass
        else:
            pass
        # Divergence Error Handling
        if i > limit:
            raise RuntimeError("Fail to Converge.")
            break
        i += 1



from uncertainties import ufloat
from uncertainties.umath import *


def printp(x, txt=None, txtxt=None, num=2):  # 이쁘게 출력해준다.
    temp = '{:.' + str(num) + 'uP}'
    if txt is None:
        txt = ''
    else:
        txt = str(txt)
    if txtxt is None:
        txtxt = ''
    else:
        txtxt = str(txtxt)
    print(txt + temp.format(x) + txtxt)


def printp_list(X, txt=None, txtxt=None, num=2):  # 리스트를 이쁘게 출력해준다. 한 줄에 한 원소.
    for x in X:
        printp(x, txt=txt, txtxt=txtxt, num=num)


def printp_list_oneline(x, txt='[', txtxt=']', num=2):  # 리스트를 이쁘게 출력해준다. 한 줄에 다 출력한다.
    n = len(x)
    temp = '{:.' + str(num) + 'uP}'

    if txt is None:
        txt = ''
    else:
        txt = str(txt)
    if txtxt is None:
        txtxt = ''
    else:
        txtxt = str(txtxt)

    print(txt, end = '')
    i = 0
    while i < n:
        if i == n-1:
            END = ''
        else:
            END = ', '
        print(temp.format(X[i]), end = END)
        i = i+1
    print(txtxt, end = '\n')


import numpy as np
import matplotlib.pyplot as plt
import matplotlib
# matplotlib.rcParams['text.usetex'] = False
    # 원래 spyder면 레이텍을 불러와야 하지만, 코랩에서는 불러올 수 없는데다가,
    # 없어도 그냥 쓸 수 있다.

import uncertainties
from uncertainties import ufloat
from uncertainties.umath import *

import os  # path 수정용
import platform  # windows 확인용

import copy  # 복사용

from exp import *
from linfit import *


UF = "<class 'uncertainties.core.Variable'>"
UFAS = "<class 'uncertainties.core.AffineScalarFunc'>"
UFlist = [UF, UFAS]


# 자동으로 좋은 plot을 짜주는 함수를 짜고 싶다.
# animation도 목표로 넣도록 하자.
    # 진짜?


class superplt():
    def write_superplt(relation, Directory, textTitle, Time=None):
        # input : relation 의 입력 검사
        if isinstance(relation, Shvar) is False:
            raise TypeError("Input 'relation' should be a Shvar class object.")

        if relation.tp != 'double group':
            raise TypeError("Input 'relation' should have .tp value as 'double group'.")

        # input : Directory 의 입력 검사
        if isinstance(Directory, str) is False:
            raise TypeError("Input 'Directory' should be a string.")

        if os.path.isdir(Directory) is False:
            raise FileNotFoundError("The Directory does not exist.")

        TEXT = []

        # 작성 시간
        renewTime = Time.replace(';', ':')
        TEXT.append("Drew Time : " + renewTime)

        # 작성 변수
        TEXT.append("names : " + str(relation.names))
        TEXT.append("grouptp : " + str(relation.grouptp))
        TEXT.append("XYdata : " + str(relation.XYdata))

        # 피팅 결과
        if relation.slope is not None:
            TEXT.append("slope : {:.16}".format(relation.slope))
            TEXT.append("intercept : {:.16}".format(relation.intercept))
            TEXT.append("Coefficient of Determination : " + str(relation.R2))

        f = open(Directory + textTitle + '.txt', 'w')
        for x in TEXT:
            f.write(x + '\n')
        f.close()


    def plot(relation, fitxnum=1001, Capsize=3, datagraphname=None, FMT='-o',
             MS=5, fitcolor=(0.1, 0.1, 0.1, 0.9), datacolor='black'):


        """
        A method creating two plot rapidly. (not showed)

        Parameter
            relation      : The 'double group' type Shvar object to draw a plot.
            fitxnum       : The number of x-axes data used to draw a fitting graph.
            Capsize       : The wides of errorbar. default=3
            datagraphname : The name of datagraph
            FMT           : The properties of plot line and marker. Refer to matplotlib.pyplot.plot
            MS            : The markersize. default=5
            fitcolor      : The color of fitting graph.
            datacolor     : The color of data graph.
        """


        # 입력 검사.
            # 입력은 반드시 Shvar 클래스의 'double group' type이여야 한다.
        if isinstance(relation, Shvar):
            if relation.tp == 'double group':
                pass
            else:
                raise TypeError("The input 'relation' should be a Shvar class and have 'double group' type.")
        else:
            raise TypeError("The input 'relation' should be a Shvar class.")

        # X, Y가 각각 uf인 경우/아닌 경우에 맞춰 에러바를 그린다.
        if relation.grouptp[0] in ['uflist', 'ufdatalist']:
            Xn = [x.n for x in relation.XYdata[0]]
            Xs = [x.s for x in relation.XYdata[0]]
            XERR = Xs
        else:
            Xn = copy(relation.XYdata[0])
            XERR = None
        if relation.grouptp[1] in ['uflist', 'ufdatalist']:
            Yn = [y.n for y in relation.XYdata[1]]
            Ys = [y.s for y in relation.XYdata[1]]
            YERR = Ys
        else:
            Yn = copy(relation.XYdata[1])
            YERR = None

        plt.errorbar(Xn, Yn, xerr=XERR, yerr=YERR, capsize=Capsize,
                     label=datagraphname, fmt=FMT, ms=MS, color=datacolor)

        if relation.slope is not None:
            def fitting_func(x):  # 피팅 결과를 보여줄 함수
                y = relation.slope.n * x + relation.intercept.n
                return y

            # 피팅의 결과는 데이터보다 앞뒤로 좀더 보여준다. (10%)
            minx = min(Xn)
            maxx = max(Xn)
            firstx = -0.1 * (maxx - minx) + minx
            endx = 0.1 * (maxx - minx) + maxx
            fitX = np.linspace(firstx, endx, fitxnum)
            fitY = [fitting_func(x) for x in fitX]
            plt.plot(fitX, fitY, label='linear regression', color=fitcolor)


    def plot_comp(relation, fitxnum=1001, Title=None, pos_leg='right',
                  Label=['normal', 'normal'], Capsize=3, FMT='-o', MS=5,
                  datagraphname=None, canvascolor='white', backcolor='white',
                  text=True, fitcolor=(0.1, 0.1, 0.1, 0.9), datacolor='black'):


        """
        A method making a plot rapidly.

        Parameter
            relation      : The 'double group' type Shvar object to draw a plot.
            fitxnum       : The number of x-axes data used to draw a fitting graph.
            Title         : Title of plot. String type. default=None
            pos_leg       : Position of legend. 'right' or 'left'. The up/downward position of legend is decided by the sign of slope.
            Label         : The name of xlabel and ylabel. [string, string].
            Capsize       : The wides of errorbar. default=3
            FMT           : The properties of plot line and marker. Refer to matplotlib.pyplot.plot
            MS            : The markersize. default=5
            datagraphname : The name of datagraph. If this is None, Legend does not created.
            canvascolor   : The color of canvas.
            backcolor     : The color of background of plot area.
            text          : Whether the text is used. bullen type. default=True
            fitcolor      : The color of fitting graph.
            datacolor     : The color of data graph.

        This method make the following files.
            plot_date_time.png
            plotinfo_date_time.txt
        """


        # 입력 검사.
            # 입력은 반드시 Shvar 클래스의 'double group' type이여야 한다.
        if isinstance(relation, Shvar):
            if relation.tp == 'double group':
                pass
            else:
                raise TypeError("The input 'relation' should be a Shvar class and have 'double group' type.")
        else:
            raise TypeError("The input 'relation' should be a Shvar class.")

            # pos_leg 는 'right'거나 'left'여야 한다.
        if pos_leg != 'right' and pos_leg != 'left':
            raise ValueError("'pos_leg' should be 'right' or 'left'.")

        # X, Y가 각각 uf인 경우/아닌 경우에 맞춰 에러바를 그린다.
        if relation.grouptp[0] in ['uflist', 'ufdatalist']:
            Xn = [x.n for x in relation.XYdata[0]]
            Xs = [x.s for x in relation.XYdata[0]]
            XERR = Xs
        else:
            Xn = copy(relation.XYdata[0])
            XERR = None
        if relation.grouptp[1] in ['uflist', 'ufdatalist']:
            Yn = [y.n for y in relation.XYdata[1]]
            Ys = [y.s for y in relation.XYdata[1]]
            YERR = Ys
        else:
            Yn = copy(relation.XYdata[1])
            YERR = None

        fig = plt.figure()  # 캔버스 생성
        fig.set_facecolor(canvascolor)  # 캔버스 색상 설정
        ax = fig.add_subplot()  # 프레임(그림 뼈대) 생성
        ax.set_facecolor(backcolor)  # 그래프 영역 배경 색상 설정

        # 에러바 그래프 생성
        ax.errorbar(Xn, Yn, xerr=XERR, yerr=YERR, capsize=Capsize,
                    label=datagraphname, fmt=FMT, ms=MS, color=datacolor)

        # relation object에 피팅값이 있는 경우
        if relation.slope is not None:
            def fitting_func(x):  # 피팅 결과를 보여줄 함수
                y = relation.slope.n * x + relation.intercept.n
                return y

            # 피팅의 결과는 데이터보다 앞뒤로 좀더 보여준다. (10%)
            minx = min(Xn)
            maxx = max(Xn)
            firstx = -0.1 * (maxx - minx) + minx
            endx = 0.1 * (maxx - minx) + maxx
            fitX = np.linspace(firstx, endx, fitxnum)
            fitY = [fitting_func(x) for x in fitX]
            ax.plot(fitX, fitY, label='linear regression', color=fitcolor)

            # 피팅 결과은 식으로도 보여주게 되는데, 텍스트로 보여준다.
                # 텍스트의 위치는 데이터의 90%, 80%, 70% 위치 또는
                # 10%, 20%, 30% 위치에서 보여주게 된다.
            minY = min(fitY)
            maxY = max(fitY)
            posx = minx + 0.65 * (maxx - minx)  # 텍스트가 오른쪽 방면에 쓰여질 때의 x좌표이다.
            if text is True:
                if relation.slope.n >= 0:  # 기울기가 양수일 때
                    if pos_leg == 'right':  # 범례가 오른쪽일 때
                        firstposy = minY + 0.9 * (maxY - minY)
                        secondposy = minY + 0.8 * (maxY - minY)
                        thirdposy = minY + 0.7 * (maxY - minY)
                        ax.text(minx, firstposy, r'$y=ax+b$', fontsize=10)
                        ax.text(minx, secondposy, 'a={:.2uP}'.format(relation.slope), fontsize=10)
                        ax.text(minx, thirdposy, 'b={:.2uP}'.format(relation.intercept), fontsize=10)
                    else:  # 범례가 왼쪽일 때
                        firstposy = minY + 0.3 * (maxY - minY)
                        secondposy = minY + 0.2 * (maxY - minY)
                        thirdposy = minY + 0.1 * (maxY - minY)
                        ax.text(posx, firstposy, r'$y=ax+b$', fontsize=10)
                        ax.text(posx, secondposy, 'a={:.2uP}'.format(relation.slope), fontsize=10)
                        ax.text(posx, thirdposy, 'b={:.2uP}'.format(relation.intercept), fontsize=10)
                else:  # 기울기가 음수일 때
                    if pos_leg == 'right':  # 범례가 오른쪽일 때
                        firstposy = minY + 0.3 * (maxY - minY)
                        secondposy = minY + 0.2 * (maxY - minY)
                        thirdposy = minY + 0.1 * (maxY - minY)
                        ax.text(minx, firstposy, r'$y=ax+b$', fontsize=10)
                        ax.text(minx, secondposy, 'a={:.2uP}'.format(relation.slope), fontsize=10)
                        ax.text(minx, thirdposy, 'b={:.2uP}'.format(relation.intercept), fontsize=10)
                    else:  # 범례가 왼쪽일 때
                        firstposy = minY + 0.9 * (maxY - minY)
                        secondposy = minY + 0.8 * (maxY - minY)
                        thirdposy = minY + 0.7 * (maxY - minY)
                        ax.text(posx, firstposy, r'$y=ax+b$', fontsize=10)
                        ax.text(posx, secondposy, 'a={:.2uP}'.format(relation.slope), fontsize=10)
                        ax.text(posx, thirdposy, 'b={:.2uP}'.format(relation.intercept), fontsize=10)

        # Xlabel 작성
        if Label[0] == 'normal':
            ax.set_xlabel(relation.names[0])
        else:
            ax.set_xlabel(Label[0])

        # Ylabel 작성
        if Label[1] == 'normal':
            ax.set_ylabel(relation.names[1])
        else:
            ax.set_ylabel(Label[1])

        # 제목 작성
        ax.set_title(Title)

        # 그리드 작성
        ax.grid()

        # 범례 작성
        if datagraphname is not None:
            handles, labels = ax.get_legend_handles_labels()  # 범례 처리되는 요소와 해당 라벨
            dict_labels_handles = dict(zip(labels, handles))  # 라벨을 키로 요소를 밸류로 하는 딕셔너리 생성
            labels = [datagraphname, 'linear regression']  # 원하는 순서 라벨
            handles = [dict_labels_handles[j] for j in labels]  # 라벨 순서에 맞게 요소 재배치
            if pos_leg == 'right':
                if relation.slope.n > 0:
                    ax.legend(handles, labels, loc='lower right')
                else:
                    ax.legend(handles, labels, loc='upper right')
            else:
                if relation.slope.n > 0:
                    ax.legend(handles, labels, loc='upper left')
                else:
                    ax.legend(handles, labels, loc='lower left')

        # fig.show()

        # 백업 파일 기본 디렉토리
        if platform.system() == 'Windows':
            BaseDirectory = __file__.replace('exp_plot.py', 'plots\\')
            if os.path.isdir(BaseDirectory.replace('plots\\', 'plots')) is False:
                os.mkdir(BaseDirectory.replace('plots\\', 'plots'))
        else:
            BaseDirectory = __file__.replace('exp_plot.py', 'plots/')
            if os.path.isdir(BaseDirectory.replace('plots/', 'plots')) is False:
                os.mkdir(BaseDirectory.replace('plots/', 'plots'))

        timelist = Shvar.change_time_to_text()  # 현재시간을 텍스트 형식으로 불러온다.
        time_for_Title = timelist[0] + '_' + timelist[1]  # 제목으로 바꾼다.
        Title = 'plot_' + time_for_Title  # 플롯 제목
        textTitle = 'plotinfo_' + time_for_Title  # 플롯 인포메이션 제목
        plt.savefig(BaseDirectory + Title + '.png')  # 플롯 저장
        superplt.write_superplt(relation, BaseDirectory, textTitle, Time=time_for_Title)



from exp import Shvar
from exp_plot import superplt
import uncertainties
from uncertainties import ufloat
from uncertainties.umath import *
import matplotlib.pyplot as plt
import numpy as np

# 이 파일은 Shvar 라이브러리의 튜토리얼입니다.
# 이 튜토리얼은 python IDE인 spyder를 기준으로 작성되었습니다.
# 혹여나 다른 IDE로 작동하시는 분은 설정에 유의해주시기 바랍니다.

# 이 튜토리얼은 각 함수마다 셀이 나뉘어져 있습니다.
    # spyder를 기준으로 ctrl+Enter를 하면 선택되어있는 셀만 실행할 수 있습니다.
    # 반드시 셀 순서를 지켜서 실행해주시기 바랍니다.

# 목차
    # 1. 세팅
    # 2. 변수설정
    # 3. 백업 & 백업 파일 불러오기
    # 4. 피팅
    # 5. plot 출력
    # 6. 변수간의 계산
    # 7. class variable 상세 내용
    

# 1. 세팅
    # exp.py와 exp_plot.py를 한 파일 안에 두어야 합니다.
    # 추가적인 세팅 없이 사용하고 싶으시다면, 다른 라이브러리가 저장되는 경로에 다른 라이브러리와 동일한 방법으로 저장해 주시면 되겠습니다.
    # 예시 : C:\Users\jang9\AppData\Local\Programs\Python\Python39\Lib\site-packages
    # 그렇지 않고 다른 파일에 해당 파일들을 저장하시는 경우, IDE의 세부설정에 들어가서 path에 추가해주시기 바랍니다.
    # 이 라이브러리를 사용할때는 이 파일이 import하고 있는 라이브러리 6개를 기본으로 import하고 시작해주시기 바랍니다.


# 2. 변수설정
    # Shvar 클래스의 오브젝트를 선언하고 싶다면 다음과 같이 코드를 작성합니다.
# %% 클래스 오브젝트 선언

A = Shvar()

# %%
    # 이렇게 되면 A는 모종의 상자처럼 작동하게 되며, 이 안에 알맞는 데이터를 넣음으로써, 자유자재로 운용할 수 있게 됩니다.
    
    # Shvar 클래스에서 사용할 수 있는 타입은 다음이 있습니다.
        # 1. data의 리스트를 넣어서 ufloat 하나를 만드는 타입 (setufdata)
        # 2. data의 리스트의 리스트를 넣어서 ufloat의 리스트를 만드는 타입 (setufdatalist)
        # 3. ufloat을 넣거나 meanvalue와 uncertainty를 넣어서 ufloat 하나를 만드는 타입 (simpleuf)
        # 4. ufloat의 리스트를 넣거나 meanvalue의 리스트와 uncertianty의 리스트를 넣어서 ufloat의 리스트를 만드는 타입 (simpleuflist)
        # 5. 수 하나를 넣는 타입 (simplescalar)
        # 6. 수의 리스트를 넣는 타입 (simplescalarlist)
        # 7. 두 개의 리스트를 묶어서 넣는 타입 (linear_rel)
    # 시험삼아 setufdata를 이용해 Shvar 클래스 변수를 생성해보겠습니다.
# %% setufdata 예시

Measure = [11.2, 10.8, 10.0, 11.5, 10.5]
A.setufdata(Measure)
print(A.uf)

# %%
    # setufdata 함수를 통해 Shvar 클래스 오브젝트(A)에 어떤 리스트를 넣어주면, 자동으로 평균과 uncertainty를 구해 저장을 해둡니다.
    # 다음은 Shvar 클래스 오브젝트(A)에 저장되어있는 자주 사용할 variable입니다.

# %% class variable 예시

print("measured value :", A.uf)
print("meanvalue :", A.mean)
print("uncertainty :", A.SE)
print("Original Data :", A.origin)

# %%
    # 이런 class variable은 다음과 같이 한꺼번에 출력하는 것이 가능합니다.

# %%

A.printfull()

# %%
    # 위 함수를 사용해 출력되는 class variable은 유의미한 값을 가진 variable만 출력됩니다.
    # 타입에 따라 출력되는 variable이 다양합니다.

    # 위에서 Measure를 측정값으로 봤을 때, 눈금에 의한 uncertainty도 고려를 해주어야 합니다.
    # 눈금 간격이 0.1이었을 때, 그로 인한 오차는 눈금간격의 1/20 이므로, 다음과 같이 설정해 주어, 눈금에 의한 uncertainty도 고려를 해줄 수 있습니다. (무시한 경우와 큰 차이는 안 날 수 있습니다.)

# %%

A.setufdata(Measure, ec=0.1 * 0.05)
A.printfull()

# %%
    # 다음은 setufdatalist, simpleuf, simpleuflist, simplescalar, simplescalarlist의 예시입니다.

# %%

print('\nsetufdatalist case')
Measurelist = [[x + i for x in Measure] for i in range(10)]
A = Shvar()
A.setufdatalist(Measurelist)
A.printfull()

print('\nsimpleuf case - 1')
Mass = ufloat(10, 0.1)
A = Shvar()
A.simpleuf(Mass)
A.printfull()

print('\nsimpleuf case - 2')
Mass_mean = 10
Mass_error = 0.1
A = Shvar()
A.simpleuf(Mass_mean, Mass_error)
A.printfull()

print('\nsimpleuflist case - 1')
Mass_list = [ufloat(10 + i, 0.1) for i in range(10)]
A = Shvar()
A.simpleuflist(Mass_list)
A.printfull()

print('\nsimpleuflist case - 2')
Mass_mean_list = range(10)
Mass_error_list = [0.1 for i in range(10)]
A = Shvar()
A.simpleuflist(Mass_mean_list, Mass_error_list)
A.printfull()

print('simplescalar case')
PI = 3.14
A = Shvar()
A.simplescalar(PI)
A.printfull()

print('\nsimplescalarlist case')
NUMBER = range(6)
A = Shvar()
A.simplescalarlist(NUMBER)
A.printfull()

# %%

# 3. 백업 & 백업 불러오기
    # 설정한 변수들의 백업 파일은 자동으로 생성되며, exp.py 와 exp_plot.py를 저장한 경로에서 var라는 파일안에 txt파일로 저장되게 됩니다.
    # 저장 시점은 변수가 생성 및 갱신 될 때이고, 보통은 덮어쓰기 되어지나, 30분 간격으로 새 백업 파일을 만들게 됩니다.
    # 백업한 데이터를 다시 불러올 수 있으며, 다음과 같이 실행합니다.

# %%

B = Shvar()
B = Shvar.readshvartext('A', imp=True)  # 반드시 imp=True를 넣어줘야 함
    # imp=False를 하게 되면, B에 아무것도 assign되지 않으며, 그저 print만 됨.
print('백업된 변수')
A.printfull()
print('백업을 불러온 변수')
B.printfull()

# %%

# 4. 피팅
    # 이 라이브러리는 linear_rel 함수를 통해 자동으로 linear regression을 해준다.
    # 주의 할 점은, uncertainty를 집어넣으면 자동으로 Weight로써 고려하게 된다.
    # linear_rel 함수에 X데이터, Y데이터로 넣을 수 있는 형식은 다음과 같다.
        # setufdatalist로 데이터를 넣은 타입
        # simplescalarlist로 데이터를 넣은 타입
        # 일반적인 리스트는 input으로 받지 못하므로, Shvar class object에 집어넣은 후에 사용해야 한다.

# %%

C = Shvar()
C.simplescalarlist([1, 2, 3, 4, 5, 6, 7, 8, 9, 10])

D = Shvar()
D.simpleuflist([Mass_list[i] + ufloat(np.sin(i), abs(np.cos(i))) for i in range(len(Mass_list))])

E = Shvar()
E.linear_rel(C, D)  # 현재 버전에서 linear_rel(C, D+1) 같은 식으로 사용할 수 없다.

E.printfull()

# %%

# 5. plot
    # 앞서 linear_rel를 이용해 선언된 object를 이용하면 바로 plot을 그릴 수 있다.

# %%

superplt.plot_comp(E, Title='test', Label=['X', 'Y'], datagraphname='testgraph', datacolor='blue')

# %%
    
    # plot_comp를 통해 생성된 plot과 그에 사용된 데이터에 대한 백업파일은 함수의 작동과 동시에 생성되며, 위치는 exp_plot.py가 저장된 폴더의 'plots'이다. plot은 이미지로, 그에 사용된 데이터는 txt로 저장된다.
    # plot_comp는 plt.figure부터 plt.show까지 한번에 이루어지기 때문에, 수정이 어렵다는 단점이 있다.
    # 수정을 거치고 싶을 경우, 특히 여러개를 한번에 그리고 싶은 경우 다음과 같이 사용하면 된다.

# %%

F = Shvar()
F.linear_rel(A, B)

plt.figure()
superplt.plot(F, datagraphname='hi')
superplt.plot(E, datacolor='blue', fitcolor='red', datagraphname='hello')
plt.title("test title")
plt.xlabel('x')
plt.ylabel('y')
plt.grid()
plt.legend()
plt.show()

# %%

# 6. 변수간의 계산
    # Shvar class로 생서된 object는 .tp라는 class variable을 갖게 된다.
        # .tp에는 일종의 '타입명'이 저장된다.
    # .tp에는 다음이 들어갈 수 있다.
        # 1. 'ufdata' : setufdata 함수로 생성된 Shvar class object의 .tp값이다.
        # 2. 'ufdatalist' : setufdatalist 함수로 생성된 Shvar class object의 .tp값이다.
        # 3. 'uf' : simpleuf 함수로 생성된 Shvar class object의 .tp값이다.
        # 4. 'uflist' : simpleuflist 함수로 생성된 Shvar class object의 .tp값이다.
        # 5. 'scalar' : simplescalar 함수로 생성된 Shvar class object의 .tp값이다.
        # 6. 'scalarlist' : simplescalarlist 함수로 생성된 Shvar class object의 .tp값이다.
        # 7. 'double group' : linear_rel 함수로 생성된 Shvar class object의 .tp값이다.
        # 8. None : object가 텅 비어있으면 가질 수 있는 .tp값이다.
    # Shvar class object는 기초적인 계산에 대한 계산을 지원하며, 연산이 가능한 .tp 조합은 다음과 같다.
        # 더하기 (+), 빼기 (-), 곱하기(*), 나누기(*)
            # ufdata + ufdata, ufdatalist, uf, uflist, scalar, scalarlist, 숫자
            # ufdatalist + ufdata, uf, scslar, 숫자
            # uf + ufdata, ufdatalist, uf, uflist, scalar, scalarlist, 숫자
            # uflist + ufdata, uf, scslar, 숫자
            # scalar + ufdata, ufdatalist, uf, uflist, scalar, scalarlist, 숫자
            # scalarlist + ufdata, uf, scslar, 숫자
            # double group은 연산이 불가능하다.
        # 위에서 언급한 숫자에는 ufloat도 들어간다. 자세한 이야기는 라이브러리 uncertainties를 참고.

# %%

uf = Shvar()
uf.simpleuf(10, 0.2)
ufdata = Shvar()
ufdata.setufdata(range(10))
uflist = Shvar()
uflist.simpleuflist([ufloat(i, 0.2 * i) for i in range(10)])
ufdatalist = Shvar()
ufdatalist.setufdatalist(Measurelist)
scalar = Shvar()
scalar.simplescalar(3.14)
scalarlist = Shvar()
scalarlist.simplescalarlist(range(10))
ufloatobject = ufloat(10, 0.2)
number = 10
realnumber = 10.4

result = uf + ufdata
print(result.uf)
result = uf + uflist
print(result.uf)
result = uf + ufdatalist
print(result.uf)

# %%

# 7. class variable 상세 내용
    # 다음은 Shvar class object가 가지는 class variable의 리스트와 그 내용이다.
        # self.tp : Shvar object의 타입명이 저장된다.
        # self.origin : ufdata와 ufdatalist 타입만 이 값이 저장되며, 변수 생성당시에 넣었던 데이터 원본이 저장된다.
        # self.eachlength : ufdatalist 타입만 이 값이 저장되며, 변수 생성당시에 넣어던 리스트의 원소들(얘도 리스트)의 길이가 리스트 형식으로 저장된다.
        # self.tbe : ufdata와 ufdatalist 타입만 이 값이 저장되며, rectangle distribution에 의해 생성된 type_b uncertainty값이 저장된다. 예시) 자의 눈금에 의한 uncertainty
        # self.mean : uf, uflist, ufdata, ufdatalist 타입만 이 값이 저장되며, uf, ufdata는 평균값, uflist, ufdatalist는 평균의 리스트가 저장된다.
        # self.vrc : ufdata, ufdatalist 타입만 이 값이 저장되며, 분산 또는 분산의 리스트가 들어간다.
        # self.stdv : ufdata, ufdatalist 타입만 이 값이 저장되며, 표준편차 또는 표준편차의 리스트가 들어간다.
        # self.SE : uf, uflist, ufdata, ufdatalist 타입만 이 값이 저장되며, 표준오차 또는 표준오차의 리스트가 들어간다.
        # self.uf : uf, uflist, ufdata, ufdatalist 타입만 이 값이 저장되며, ufloat(평균, 표준오차) 또는 그 리스트가 저장된다.
        # self.iftval : ufdata, ufdatalist 타입만 이 값이 저장되며, 실제값이 있는지 없는지 boolean의 형태로 저장된다.
        # self.tval : ufdata, ufdatalist 타입만 이 값이 저장되며, 실제값이 있는 경우에 한하여 그 실제값을 저장한다.
        # self.length : ufdata, ufdatalist, uflist, scalarlist 타입만 이 값이 저장되며, 리스트의 길이가 저장된다.
        # self.dsc : write_des 함수를 통해 적은 description이 여기에 들어간다.
        # self.scalar : scalar 타입만 이 값이 저장되며, scalar 값이 저장된다.
        # self.scalarlist : scalarlist 타입만 이 값이 저장되며, scalarlist 값이 저장된다.
        # self.intercept : double group 타입만 이 값이 저장되며, 피팅으로 나온 y절편 값을 저장한다.
        # self.slope : double group 타입만 이 값이 저장되며, 피팅으로 나온 기울기 값을 저장한다.
        # self.R2 : double group 타입만 이 값이 저장되며, 피팅으로 나온 R^2 값을 저장한다.
        # self.names : double group 타입만 이 값이 저장되며, X, Y 데이터의 varname을 리스트로 저장한다.
        # self.grouptp : double group 타입만 이 값이 저장되며, X, Y 데이터의 tp를 리스트로 저장한다.
        # self.XYdata : double group 타입만 이 값이 저장되며, X, Y 데이터의 값을 리스트로 저장한다.
        # self.varname : 모든 타입에 대해서 저장되며, 변수명이 저장된다.



from setuptools import setup
import setuptools

setup(
      name='EXP_jangddol',
      version='0.0.2',
      url='https://github.com/jangddol/EXP_jangddol_pip.git',
      description='kawaii experiment data analysis tool',
      author='jangddol',
      author_email='j991005@korea.ac.kr',
      packages=setuptools.find_packages(),
      install_requires=[
          'uncertainties==3.1.6',
          'matplotlib==3.5.0',
          'numpy==1.21.0'
      ],
      classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent"
      ],
)


import numpy as np

import uncertainties
from uncertainties import ufloat
from uncertainties.umath import *

import os
import platform

import datetime
import pytz

import copy
import traceback

import inspect
from RIC import RememberInstanceCreationInfo

import linfit
import uprint


UF = "<class 'uncertainties.core.Variable'>"
UFAS = "<class 'uncertainties.core.AffineScalarFunc'>"
UFlist = [UF, UFAS]


def extract_varname(SELF):
    for frame, line in traceback.walk_stack(None):
        varnames = frame.f_code.co_varnames
        if varnames is ():
            break
        if frame.f_locals[varnames[0]] not in (SELF, SELF.__class__):
            break
            # if the frame is inside a method of this instance,
            # the first argument usually contains either the instance or
            #  its class
            # we want to find the first frame, where this is not the case
    else:
        pass
        # raise InstanceCreationError("No suitable outer frame found.")
    SELF._outer_frame = frame
    SELF.creation_module = frame.f_globals["__name__"]
    SELF.creation_text = traceback.extract_stack(frame, 1)[0][3]
    SELF.creation_name = SELF.creation_text.split("=")[0].strip()
    return SELF.creation_name


class Shvar(RememberInstanceCreationInfo):
    ClassVarList = ['tp', 'origin', 'eachlength', 'tbe', 'mean', 'vrc', 'stdv',
                    'SE', 'uf', 'iftval', 'tval', 'length', 'dsc', 'scalar',
                    'scalarlist', 'intercept', 'slope', 'R2', 'names',
                    'grouptp', 'XYdata', 'varname']


    def __init__(self):
        super().__init__()

        # Shvar Variable이 가지는 기본적인 저장값.
        self.tp = None
        self.origin = None
        self.eachlength = None
        self.tbe = None
        self.mean = None
        self.vrc = None
        self.stdv = None
        self.SE = None
        self.uf = None
        self.iftval = None
        self.tval = None
        self.length = None
        self.dsc = None
        self.scalar = None
        self.scalarlist = None
        self.intercept = None
        self.slope = None
        self.R2 = None
        self.names = None
        self.grouptp = None
        self.XYdata = None
        self.varname = self.creation_name
        self.writeshvartext()
    
    def nonwrite_init(self):
        self.tp = None
        self.origin = None
        self.eachlength = None
        self.tbe = None
        self.mean = None
        self.vrc = None
        self.stdv = None
        self.SE = None
        self.uf = None
        self.iftval = None
        self.tval = None
        self.length = None
        self.dsc = None
        self.scalar = None
        self.scalarlist = None
        self.intercept = None
        self.slope = None
        self.R2 = None
        self.names = None
        self.grouptp = None
        self.XYdata = None

    def setufdata(self, data, epdf='rec', ec=0, iftval=False, tval=0):
        """
        A method converting A list of number(data) to ufloat(mean+-SE).

        Parameter
            data   : A list of numbers. datalist.
            epdf   : Error Probability Distribution Funtion
                Available epdf : 'rec'
            ec     : Error Coefficient ('rec' -> Error bound)
            iftval : Whether True Value needed.
            tval   : True Value.

        This method produce following class variable.
            .tp         : ='ufdata'
            .origin     : Original data. =data
            .length     : =len(data)
            .tbe        : Type-B uncertainty
            .mean       : The average of data.
            .vrc        : The variance of data. if iftval==True, meaningless.
            .stdv       : The standard deviation of data.
            .SE         : The standard Error of data.
            .uf         : =ufloat(self.mean, self.SE)
            .iftval     : Whether True Value needed.
            .tval       : True Value.

        This method return .uf.
        """
        
        # Error Handling
        try:
            data[0]
        except TypeError:
            raise TypeError("Input 'data' should be subscriptable.")
        Shvar.nonwrite_init(self)
        if epdf == 'rec':  # Error Probability Distribution Function Checking
            if isinstance(ec, (int, float)) and ec >= 0:
                self.tbe = ec / np.sqrt(3)
            else:
                raise TypeError("Parameter 'ec' should be a positive number.")
        else:
            raise TypeError("%s is not available 'epdf'."%(epdf))
        if isinstance(tval, (int, float)) is False:
            raise TypeError("Parameter 'tval' should be a number.")
        N = len(data)
        self.tp = 'ufdata'
        self.origin = list(data)
        self.length = N
        self.mean = np.average(data)
        if iftval is True:  # If data has true value, then the degree of freedom is 'n'.
            self.tval = tval
            self.vrc = sum([(data[i] - tval)**2 for i in range(N)]) / N
        elif iftval is False:  # If data has no true value, then the degree of freedom is 'n-1'.
            self.vrc = sum([(data[i] - self.mean)**2 for i in range(N)]) / (N - 1)
        else:
            raise TypeError("Parameter 'iftval' should be True or False.")
        self.stdv = np.sqrt(self.vrc)
        self.SE = np.sqrt((self.stdv**2) / N + self.tbe**2)  # Propagation of Uncertainties.
        self.uf = ufloat(self.mean, self.SE)
        self.writeshvartext()
        return self


    def setufdatalist(self, data, epdf='rec', ec=0, iftval=False, tval=0):
        """
        A method converting A list of list of number(data) to A list of ufloat(mean+-SE).

        Parameter
            data   : A list of list of numbers. list of datalist.
            epdf   : Error Probability Distribution Funtion
                Available epdf : 'rec'
            ec     : Error Coefficient ('rec' -> Error bound)
            iftval : Whether True Value needed.
            tval   : True Value.
        
        This method produce following class variable.
            .tp         : ='ufdatalist'
            .origin     : Original data. =data
            .eachlength : The list of length of data which was used to calculate each ufdata.
            .tbe        : Type-B uncertainty
            .mean       : The list of average of data.
            .vrc        : The list of variance of data. if iftval==True, meaningless.
            .stdv       : The list of standard deviation of data.
            .SE         : The list of standard Error of data.
            .uf         : The list of ufloat(self.mean, self.SE)
            .iftval     : Whether True Value needed.
            .tval       : The list of True Value.
            .length     : The number of uf value.
        
        This method return .uf.
        """
        
        # Error Handling
        try:
            data[0]
        except TypeError:
            raise TypeError("Input 'data' should be subscriptable.")
        
        N = len(data)
        for X in data:
            try:
                X[0]
            except TypeError:
                raise TypeError("The elements of Input 'data' should be subscriptable.")
        
        for X in data:
            for y in X:
                if isinstance(y, (int, float)) is False:
                    raise TypeError("Input 'data' should be a list of list of numbers.")
        
        Shvar.nonwrite_init(self)
        
        if epdf == 'rec':  # Error Probability Distribution Function Check
            if isinstance(ec, (int, float)) and ec >= 0:
                self.tbe = ec / np.sqrt(3)
            else:
                raise TypeError("Parameter 'ec' should be a positive number.")
        else:
            raise TypeError("%s is not available 'epdf'."%(epdf))
        
        self.tp = 'ufdatalist'
        self.origin = list([list(x) for x in data])
        self.eachlength = [len(data[i]) for i in range(N)]
        self.mean = [np.average(data[i]) for i in range(N)]
        
        if iftval is True:  # If data has True Value, the degree of freedom is 'n'.
            # tval should be a list of numbers which have same length with 'data'.
            try:
                tval[0]
            except TypeError:
                raise TypeError("If iftval is True, tval should be a list of numbers which have same length with 'data'")
            
            if len(tval) != N:
                raise TypeError("If iftval is True, 'data' and 'tval' should have same length.")
            
            for x in tval:
                if isinstance(x, (int, float)) is False:
                    raise TypeError("If iftval is True, 'tval' should be a list of numbers.")
            
            self.vrc = [sum([(data[i][j] - tval[i])**2 for j in range(len(data[i]))]) / self.eachlength[i] for i in range(N)]
            self.tval = tval
        elif iftval is False:  # If data has no True Value, the degree of freedom is 'n-1'.
            self.vrc = [sum([(data[i][j] - self.mean[i])**2 for j in range(len(data[i]))]) / (self.eachlength[i] - 1) for i in range(N)]
        else:
            raise TypeError("Parameter 'iftval' should be True or False.")
        self.stdv = [np.sqrt(self.vrc[i]) for i in range(N)]
        self.SE = [np.sqrt((self.stdv[i]**2) / self.eachlength[i] + self.tbe**2) for i in range(N)]
        self.uf = [ufloat(self.mean[i], self.SE[i]) for i in range(N)]
        self.length = N
        self.writeshvartext()
        return self


    def write_des(self, description):
        """
        A method making a description to variable.

        Parameter
            description : A string to be a description of self.
        """
        
        if (type(description) is str) is False:
            raise TypeError("The description should be a string.")
        self.dsc = description
        self.writeshvartext()
    

    def simplescalar(self, scalar):
        """
        A method making a scalar.

        Parameter
            scalar : A scalar to generate
        
        This method makes the following variables.
            .scalar : = scalar
            .tp     : 'scalar'
        """
        
        if isinstance(scalar, (int, float)) is False:
            raise TypeError("'scalar' should be a number.")
        Shvar.nonwrite_init(self)
        self.scalar = scalar
        self.tp = 'scalar'
        self.writeshvartext()
        return self
    

    def simplescalarlist(self, scalarlist):
        """
        A method making a scalarlist.

        Parameter
            scalarlist : A list of numbers.
        
        This method make the following values.
            .length     : The length of scalarlist
            .tp         : 'scalarlist'
            .scalarlist : The list of numbers
        
        This method return the scalarlist.
        """
        
        # Error Handling
        try:
            scalarlist[0]
        except Exception:
            raise TypeError("The scalarlist should be subscriptable.")

        for scalar in scalarlist:
            if isinstance(scalar, (int, float)) is False:
                raise TypeError("The scalarlist should be a list of numbers.")
                break
        Shvar.nonwrite_init(self)
        self.length = len(scalarlist)
        self.tp = 'scalarlist'
        self.scalarlist = list(scalarlist)
        self.writeshvartext()
        return self
    

    def simpleuf(self, *args):
        """
        A method making an ufloat type variable.
        
        Parameter
            If args is one, the argument should be an ufloat value
            If args are two, the arguments should be two numbers.
                These numbers would be the mean and the standard error values.
        
        This method makes the following variables.
            .mean : A mean value.
            .SE   : A standard error value.
            .uf   : An ufloat value.
            .tp   : 'uf'
        """
        
        # 오리지널 데이터가 없어도 클래스 변수를 입력하는 메서드
        # 숫자 2개가 들어오면 uf를 반환
        # uf를 입력받으면 평균과 표준편차를 반환
        Shvar.nonwrite_init(self)
        if len(args) == 1:
            if str(type(args[0])) not in UFlist:
                raise TypeError("The argument should be an ufloat value.")
            else:
                a = args[0]
                self.mean = a.n
                self.SE = a.s
                self.uf = a
                self.tp = 'uf'
        elif len(args) == 2:
            if isinstance(args[0], (int, float)) and isinstance(args[1], (int, float)):
                self.mean = args[0]
                self.SE = args[1]
                self.uf = ufloat(self.mean, self.SE)
                self.tp = 'uf'
            else:
                raise TypeError("The arguments should be two numbers.")
        else:
            raise TypeError("The number of arguments should be one or two.")
        self.writeshvartext()
        return self
    
    
    def simpleuflist(self, *args):
        """
        A method making an ufloat type variable.
        
        Parameter
            If args is one, the argument should be a list of ufloat value
            If args are two, the arguments should be two lists of numbers.
                These lists would be the list of means and the list of standard errors.
        
        This method makes the following variables.
            .mean   : The list of mean value.
            .SE     : The list of standard error value.
            .uf     : The list of ufloat value.
            .tp     : 'uf'
            .length : The length of uf list.
        """

        # Error Handling
        for a in args:
            try:
                a[0]
            except TypeError:
                raise TypeError("The arguments should be one or two lists.")
        Shvar.nonwrite_init(self)
        if len(args) == 1:
            for x in args[0]:
                if str(type(x)) not in UFlist:
                    raise TypeError("The argument should be a list of ufloats.")
            self.uf = args[0]
            self.mean = [x.n for x in args[0]]
            self.SE = [x.s for x in args[0]]
            self.tp = 'uflist'
        elif len(args) == 2:
            if len(args[0]) != len(args[1]):
                raise TypeError("The two arguments should have same length.")
            for x in args[0]:
                if isinstance(x, (int, float)) is False:
                    raise TypeError("The first argument(mean) should be a list of numbers.")
            for y in args[1]:
                if (isinstance(y, (int, float)) and y > 0) is False:
                    raise TypeError("The second argument(SE) should be a list of positive numbers.")
            self.uf = [ufloat(args[0][i], args[1][i]) for i in range(len(args[0]))]
            self.mean = args[0]
            self.SE = args[1]
            self.tp = 'uflist'
            self.length = len(self.uf)
        else:
            raise TypeError("The number of arguments should be one or two.")
        self.writeshvartext()
        return self


    def __add__(self, other):
        # 'ufdatalist', 'uflist', 'ufdata', 'uf', 'scalar', 'scalarlist'의 경우를 모두 고려해야한다.
        # 'ufdatalist', 'uflist', 'scalarlist' 를 'list',
        # 'ufdata', 'uf', 'scalar' 를 'uf'로 단순화할 때,
            # 'list' + 'uf', 'uf' + 'list', 'list' + num, 'uf' + 'uf', 'uf' + num
            # 가 가능한 케이스다.
        # Shvar가 아닌 경우는 int, float, ufloat을 고려한다.
        self.creation_name = extract_varname(self)
        
        result = Shvar()
        mode = 0
        if self.tp == 'ufdatalist' or self.tp == 'uflist':
            mode = 'uflist'
        elif self.tp == 'ufdata' or self.tp == 'uf':
            mode = 'uf'
        elif self.tp == 'scalarlist':
            mode = 'slist'
        elif self.tp == 'scalar':
            mode = 's'
        else:
            print(self.tp)
            raise TypeError("Non-addable type : tp = " + self.tp)
        
        if mode == 'uflist':
            if type(other) == type(result):  # 'uflist' + Shvar
                if other.tp == 'uf' or other.tp == 'ufdata':
                    result.simpleuflist([x + other.uf for x in self.uf])
                elif other.tp == 'scalar':
                    result.simpleuflist([x + other.scalar for x in self.uf])
                else:
                    raise TypeError("Non-addable type : " + self.tp + ' + ' + other.tp)
            else:  # 'uflist' + ufloat, 'uflist' + float
                result.simpleuflist([x + other for x in self.uf])
        elif mode == 'uf':
            if type(other) == type(result):  # 'uf' + Shvar
                if other.tp == 'uf' or other.tp == 'ufdata':
                    result.simpleuf(self.uf + other.uf)
                elif other.tp == 'scalar':
                    result.simpleuf(self.uf + other.scalar)
                elif other.tp == 'uflist' or other.tp == 'ufdatalist' or 'scalarlist':
                    result = other + self
                else:
                    raise TypeError("Non-addable type : " + self.tp + ' + ' + other.tp)
            else:  # 'uf' + ufloat, 'uf' + float
                result.simpleuf(self.uf + other)
        elif mode == 'slist':
            if type(other) == type(result):
                if other.tp == 'uf' or other.tp == 'ufdata':
                    result.simpleuflist([x + other.uf for x in self.scalarlist])
                elif other.tp == 'scalar':
                    result.simplescalarlist([x + other.scalar for x in self.scalarlist])
                else:
                    raise TypeError("Non-addable type : " + self.tp + ' + ' + other.tp)
            elif str(type(other)) in UFlist:
                result.simpleuflist([x + other for x in self.scalarlist])
            else:
                result.simplescalarlist([x + other for x in self.scalarlist])
        elif mode == 's':
            if type(other) == type(result):
                result = other + self
            elif str(type(other)) in UFlist:
                result.simpleuf(self.scalar + other)
            else:
                result.simplescalar(self.scalar + other)
        else:
            raise TypeError("Something is Wrong. fxxk.")
        result.varname = self.creation_name
        result.writeshvartext()
        return result


    def __radd__(self, other):
        result = self + other
        result.writeshvartext()
        return result
    

    def __pos__(self):
        self.creation_name = extract_varname(self)

        List = ['uf', 'ufdata', 'uflist', 'ufdatalist', 'scalarlist', 'scalar']
        if self.tp in List:
            self.varname = self.creation_name
            self.writeshvartext()
            return self
        else:
            raise TypeError("Non-plus-attachable : tp = " + self.tp)


    def __neg__(self):
        self.creation_name = extract_varname(self)
        
        List = ['uf', 'ufdata', 'scalar', 'uflist', 'ufdatalist', 'scalarlist']
        if (self.tp in List) is False:
            raise TypeError("Non-minus-attachable : tp = " + self.tp)
        
        result = Shvar()
        if self.tp == 'uf' or self.tp == 'ufdata':
            result.simpleuf(-self.uf)
        elif self.tp == 'scalar':
            result.simplescalar(-self.scalar)
        elif self.tp == 'uflist' or self.tp == 'ufdatalist':
            result.simpleuflist([-x for x in self.uf])
        elif self.tp == 'scalarlist':
            result.simplescalarlist([-x for x in self.scalarlist])
        else:
            raise ValueError("Something is Wrong, fxxk.")
        
        result.varname = self.creation_name
        result.writeshvartext()
        return result
    

    def __sub__(self, other):
        self.creation_name = extract_varname(self)

        result = self + other.__neg__()
        result.varname = self.creation_name
        result.writeshvartext()
        return result
    

    def __rsub__(self, other):
        result = -(self - other)
        result.writeshvartext()
        return result
    

    def __mul__(self, other):
        self.creation_name = extract_varname(self)

        result = Shvar()
        mode = 0
        if self.tp == 'uf' or self.tp == 'ufdata':
            mode = 'uf'
        elif self.tp == 'scalar':
            mode = 's'
        elif self.tp == 'uflist' or self.tp == 'ufdatalist':
            mode = 'uflist'
        elif self.tp == 'scalarlist':
            mode = 'slist'
        else:
            raise TypeError("Non-multiplable type : tp = " + self.tp)
        
        if mode == 'uf':
            if type(other) == type(result):
                if other.tp == 'uf' or other.tp == 'ufdata':
                    result.simpleuf(self.uf * other.uf)
                elif other.tp == 'scalar':
                    result.simpleuf(self.uf * other.scalar)
                elif other.tp == 'uflist' or other.tp == 'ufdatalist':
                    result.simpleuflist([self.uf * x for x in other.uf])
                elif other.tp == 'scalarlist':
                    result.simpleuflist([self.uf * x for x in other.scalarlist])
                else:
                    raise TypeError("Non-multiplable type : " + self.tp + ' * ' + other.tp)
            else:
                result.simpleuf(self.uf * other)
        elif mode == 'uflist':
            if type(other) == type(result):
                if other.tp == 'uf' or other.tp == 'ufdata':
                    result.simpleuflist([x * other.uf for x in self.uf])
                elif other.tp == 'scalar':
                    result.simpleuflist([x * other.scalar for x in self.uf])
                else:
                    raise TypeError("Non-multiplable type : " + self.tp + ' * ' + other.tp)
            else:
                result.simpleuflist([x * other for x in self.uf])
        elif mode == 's':
            if type(other) == type(result):
                if other.tp == 'uf' or other.tp == 'ufdata':
                    result.simpleuf(self.scalar * other.uf)
                elif other.tp == 'scalar':
                    result.simplescalar(self.scalar * other.scalar)
                elif other.tp == 'uflist' or other.tp == 'ufdatalist':
                    result.simpleuflist([self.scalar * x for x in other.uf])
                elif other.tp == 'scalarlist':
                    result.simplescalarlist([self.scalar * x for x in other.scalarlist])
                else:
                    raise TypeError("Non-multiplable type : " + self.tp + ' * ' + other.tp)
            elif str(type(other)) in UFlist:
                result.simpleuf(self.scalar * other)
            else:
                result.simplescalar(self.scalar * other)
        elif mode == 'slist':
            if type(other) == type(result):
                if other.tp == 'uf' or other.tp == 'ufdata':
                    result.simpleuflist([x * other.uf for x in self.scalarlist])
                elif other.tp == 'scalar':
                    result.simplescalarlist([x * other.scalar for x in self.scalarlist])
                else:
                    raise TypeError("Non-multiplable type : " + self.tp + ' * ' + other.tp)
            elif str(type(other)) in UFlist:
                result.simpleuflist([x * other for x in self.scalarlist])
            else:
                result.simplescalarlist([x * other for x in self.scalarlist])
        else:
            raise ValueError("Something is Wrong. fxxk.")
        
        result.varname = self.creation_name
        result.writeshvartext()
        return result


    def reciprocal(self):
        self.creation_name = extract_varname(self)
        
        List = ['uf', 'ufdata', 'scalar', 'uflist', 'ufdatalist', 'scalarlist']
        if (self.tp in List) is False:
            raise TypeError("Non-calculable type(reciprocal) : tp = " + self.tp)
        
        result = Shvar()
        if self.tp == 'uf' or self.tp == 'ufdata':
            result.simpleuf(1 / self.uf)
        elif self.tp == 'scalar':
            result.simplescalar(1 / self.scalar)
        elif self.tp == 'uflist' or self.tp == 'ufdatalist':
            result.simpleuflist([1 / x for x in self.uf])
        elif self.tp == 'scalarlist':
            result.simplescalarlist([1 / x for x in self.scalarlist])
        else:
            raise ValueError("Something is Wrong, fxxk.")

        result.varname = self.creation_name
        result.writeshvartext()
        return result


    def __truediv__(self, other):
        self.creation_name = extract_varname(self)
        
        result = Shvar()
        if type(other) == type(result):
            result = self * other.reciprocal()
        else:
            result = self * (1 / other)
        result.varname = self.creation_name
        result.writeshvartext()
        return result


    def __rtruediv__(self, other):
        self.creation_name = extract_varname(self)
        
        result = Shvar()
        if type(other) == type(result):
            result = other / self
        else:
            result = other * self.reciprocal()
        result.varname = self.creation_name
        result.writeshvartext()
        return result
    

    def linear_rel(self, X, Y, fit=True):
        """
        A method group two lists with linear fitting.
        If both X and Y have uncertainties, then this method use the method of York(1966).

        Parameter
            X   : A list of numbers or a list of ufloats. The data of x-axis.
            Y   : A list of numbers or a list of ufloats. The data of y-axis.
            fit : If True, this method calculate the fitting graph
        
        This method make the following value.
            .tp : 'double group'
            .grouptp : [type of X, type of Y]
            .names : [name of X, name of Y]
            .XYdata : [value of X, value of Y]

        If fit==True, this method make the following values.
            .slope     : The slope of fitting graph (ufloat)
            .intercept : The intercept of y-axis of fitting graph (ufloat)
            .R2        : The coefficient of Determination of fitting graph (float)
        """

        # Error Handling
        if (type(X) == type(self) and type(Y) == type(self)) is False:
            raise TypeError("Input X and Y should be objects of class Shvar.")
        if fit is True and fit is False:
            raise TypeError("Parameter fit should be True or False.")
        
        Shvar.nonwrite_init(self)
        if (X.tp == 'uflist' or X.tp == 'ufdatalist') and (Y.tp == 'uflist' or Y.tp == 'ufdatalist'):
            self.grouptp = [X.tp, Y.tp]
            self.XYdata = [X.uf, Y.uf]
            if fit:
                temp = linfit.linear_fitting_York(X.uf, Y.uf, initguess='yerr')
                self.slope = ufloat(temp[0], temp[4])  # ufloat
                self.intercept = ufloat(temp[1], temp[5])  # ufloat
                self.R2 = temp[6]  # float
        elif X.tp == 'scalarlist' and (Y.tp == 'uflist' or Y.tp == 'ufdatalist'):
            self.grouptp = [X.tp, Y.tp]
            self.XYdata = [X.scalarlist, Y.uf]
            if fit:
                temp = linfit.linear_fitting_Yerr(X.scalarlist, Y.mean, Y.SE)
                self.slope = temp[0]  # ufloat
                self.intercept = temp[1]  # ufloat
                self.R2 = temp[2]  # float
        elif (X.tp == 'uflist' or X.tp == 'ufdatalist') and Y.tp == 'scalarlist':
            self.grouptp = [X.tp, Y.tp]
            self.XYdata = [X.uf, Y.scalarlist]
            if fit:
                temp = linfit.linear_fitting_Yerr(Y.scalarlist, X.mean, X.SE)
                self.slope = 1 / temp[0]  # ufloat
                self.intercept = -temp[1] / temp[0]  # ufloat
                self.R2 = temp[2]  # float
        elif X.tp == 'scalarlist' and Y.tp == 'scalarlist':
            self.grouptp = [X.tp, Y.tp]
            self.XYdata = [X.scalarlist, Y.scalarlist]
            if fit:
                temp = linfit.linear_fitting(X.scalarlist, Y.scalarlist)
                self.slope = temp[0]  # ufloat
                self.intercept = temp[1]  # ufloat
                self.R2 = temp[2]  # float
        else:
                raise TypeError("Type value of Input X and Y should be 'uflist', 'ufdatalist' or 'scalarlist'.")
        self.tp = 'double group'
        self.names = [X.varname, Y.varname]
        self.writeshvartext()
        return self
    

    def printfull(self):
        """
        This Method print the information of Shvar Variable.
        """

        print("varname :", self.varname)
        print("tp :", self.tp)
        if self.dsc is None:
            pass
        else:
            print("Description :", self.dsc)
        if self.tp == 'uf':
            print("Mean Value :", self.mean)
            print("Standard Error :", self.SE)
            uprint.printp(self.uf, txt="uf : ")
        elif self.tp == 'ufdata':
            print("Type B Error :", self.tbe)
            print("Mean Value :", self.mean)
            if self.iftval is False:
                print("Variance :", self.vrc)
                print("Standard Deviation :", self.stdv)
            print("Standard Error :", self.SE)
            uprint.printp(self.uf, txt="uf : ")
            if self.iftval:
                print("True Value :", self.tval)
            print("Origin data :", self.origin)
            print("Data Number :", self.length)
        elif self.tp == 'uflist':
            print("Length :", self.length)
            print("Mean Value :", self.mean)
            print("Standard Error :", self.SE)
            print("uf :", self.uf)
        elif self.tp == 'ufdatalist':
            print("Length :", self.length)
            print("Type B Error :", self.tbe)
            print("Mean Value :", self.mean)
            if self.iftval is False:
                print("Variance :", self.vrc)
                print("Standard Deviation :", self.stdv)
            print("Standard Error :", self.SE)
            print("uf :", self.uf)
            if self.iftval:
                print("True Value :", self.tval)
            print("Origin data :", self.origin)
            print("List of Data Number :", self.eachlength)
        elif self.tp == 'scalar':
            print("scalar :", self.scalar)
        elif self.tp == 'scalarlist':
            print("Length :", self.length)
            print("The list of scalar :", self.scalarlist)
        elif self.tp == 'double group':
            print("names :", self.names)
            print("grouptp :", self.grouptp)
            print("XYdata :", self.XYdata)
            if self.slope is not None:
                uprint.printp(self.slope, txt="slope : ")
                uprint.printp(self.intercept, txt="intercept : ")
                print("Coefficient of Detemination :", '{:.3f}'.format(self.R2))
        elif self.tp is None:
            pass
        else:
            raise TypeError("Hey, the developer need to fix this method!")
    

    def change_one_to_two_digits(inputstr):
        """
        The input 'inputstr' is a string with length 1 or 2.
        This method change the string with length 1 or 2 to one with length 2.
        """

        if isinstance(inputstr, str) is False:
            raise TypeError("inputstr should be a string.")
        
        if len(inputstr) == 1:
            return '0' + inputstr
        elif len(inputstr) == 2:
            return inputstr
        else:
            return TypeError("inputstr should be a string of length 1 or 2.")


    def change_time_to_text(deltime=0, timedeltype='past'):
        """
        This method change a time value to some following text:
            Date_text : Year-Month-Day
            time_text : Hour:Minute:Second
            Titletime_text : Hour:Minute(30minutes unit):00
        
        The initial time value is current time.
        The parameter 'deltime' means the difference of choosed time which you want and current time.
        When non-zero deltime put to input, the time value is calculated from current time.

        The output of this method is [Date_text, time_text, Titletime_text].
        """

        if (isinstance(deltime, int) and deltime >= 0) is False:  # 매개변수 deltime은 양의 정수여야 한다.
            raise TypeError("deltime should be positive integer.")
        
        if (timedeltype == 'past' or timedeltype == 'future') is False:
            # timedeltype이 'past'면 현재시간에서 deltime만큼 뺀 시간을 구한다.
            # timedeltype이 'future'면 현재시간에서 deltime만큼 더한 시간을 구한다.
            raise TypeError("timedeltype should be 'past' or 'future'.")
        
        if deltime != 0:  # 만약 deltime에 0이 아닌 정수를 입력한 경우
            len_delint = len(str(deltime))
            if len_delint <= 8:
                delzerosnum = 8 - len_delint
            else:
                raise TypeError("deltime should be 8 or less digits nubmer.")

            i = 0
            delzeros = ''
            while i < delzerosnum:
                delzeros = delzeros + '0'
                i = i + 1
            deltime_text = delzeros + str(deltime)  # 강제로 deltime을 8자리 정수인 string으로 바꿔준다.
            delDay = deltime_text[:2]  # 그후 string을 잘라서 각 시간값을 뽑아낸다.
            delHour = deltime_text[2:4]
            delMinute = deltime_text[4:6]
            delSecond = deltime_text[6:]

            current_time = datetime.datetime.now(pytz.timezone('Asia/Seoul'))
            timedelta = datetime.timedelta(days=int(delDay), seconds=int(delSecond), minutes=int(delMinute), hours=int(delHour))
            # 지정된 시간값을 뽑아낸다.

            if timedeltype == 'past':
                Time = current_time - timedelta
            else:
                Time = current_time + timedelta
        
        else:  # deltime에 0이 입력된 경우
            Time = datetime.datetime.now(pytz.timezone('Asia/Seoul'))  # 현재시간을 시간값으로 한다.
        
        Year = str(Time.year)
        Month = Shvar.change_one_to_two_digits(str(Time.month))
        Day = Shvar.change_one_to_two_digits(str(Time.day))
        Hour = Shvar.change_one_to_two_digits(str(Time.hour))
        Minute = Shvar.change_one_to_two_digits(str(Time.minute))
        Second = Shvar.change_one_to_two_digits(str(Time.second))

        Date_text = Year + '-' + Month + '-' + Day
        TitleMinute = Shvar.change_one_to_two_digits(str(30 * int(int(Minute) / 30)))
        time_text = Hour + ';' + Minute + ';' + Second  # 그냥 현재 시간
        Titletime_text = Hour + ';' + TitleMinute + ';' + '00'  # 30분 단위의 현재시간
        return [Date_text, time_text, Titletime_text]


    def check_leap_year(Year):
        """
        This method check if the input 'Year' is a leap year or not.

        The input 'Year' should be an integer.
        The output is a boolean variable.
        """

        if isinstance(Year, int) is False:
            raise TypeError("Year should be an integer.")
        
        if Year % 4 == 0:
            if Year % 100 == 0:
                if Year % 400 == 0:
                    return True
                else:
                    return False
            else:
                return True
        else:
            return False


    def split_Filename(Filename):
        """
        This method split the Filename to varname, Hour, Minute value.

        The input 'Filename' has the form 'varname_Hour:Minute:00'.
        The output is the form '[varname, Hour, Minute]'.
        """

        if isinstance(Filename, str) is False:
            raise TypeError("Filename should be a string.")
        
        varname, time = Filename.split('_')
        Hour, Minute, Second = time.split(';')
        return [varname, Hour, Minute]
    

    def split_Content(Content):
        if isinstance(Content, str) is False:
            raise TypeError("Content should be a string.")

        Content = Content.replace('\n', '')
        Content = Content.split('; ')

        if Content[0] == 'origin':
            Content[1] = Content[1].replace(' ', '')
            Content[1] = Content[1][1:-1]
            if Content[1][0] == '[':
                Content[1] = Content[1][1:-1]
                Content[1] = Content[1].split('],[')
                i = 0
                while i < len(Content[1]):
                    Content[1][i] = Content[1][i].split(',')
                    j = 0
                    while j < len(Content[1][i]):
                        Content[1][i][j] = float(Content[1][i][j])
                        j = j + 1
                    i = i + 1
            else:
                Content[1] = Content[1].split(',')
                i = 0
                while i < len(Content[1]):
                    Content[1][i] = float(Content[1][i])
                    i = i + 1
        elif Content[0] == 'eachlength':
            if Content[1][0] == '[':
                Content[1] = Content[1][1:-1]
                Content[1] = Content[1].split(',')
                i = 0
                while i < len(Content[1]):
                    Content[1][i] = int(Content[1][i])
                    i = i + 1
            else:
                Content[1] = int(Content[1])
        elif Content[0] in ['tbe', 'scalar', 'R2']:
            Content[1] = float(Content[1])
        elif Content[0] in ['mean', 'vrc', 'stdv', 'SE', 'tval']:
            if Content[1][0] == '[':
                Content[1] = Content[1][1:-1]
                Content[1] = Content[1].split(',')
                i = 0
                while i < len(Content[1]):
                    Content[1][i] = float(Content[1][i])
                    i = i + 1
            else:
                Content[1] = float(Content[1])
        elif Content[0] == 'uf':
            if Content[1][0] == '[':
                Content[1] = Content[1][1:-1]
                Content[1] = Content[1].split(',')
                i = 0
                while i < len(Content[1]):
                    x = Content[1][i].split('+/-')
                    Content[1][i] = ufloat(float(x[0]), float(x[1]))
                    i = i + 1
            else:
                Content[1] = Content[1].split('+/-')
                Content[1] = ufloat(float(Content[1][0]), float(Content[1][1]))
        elif Content[0] == 'iftval':
            if Content[1] == 'True':
                Content[1] = True
            else:
                Content[1] = False
        elif Content[0] == 'length':
            Content[1] = int(Content[1])
        elif Content[0] == 'scalarlist':
            Content[1] = Content[1][1:-1]
            Content[1] = Content[1].split(',')
            i = 0
            while i < len(Content[1]):
                Content[1][i] = float(Content[1][i])
                i = i + 1
        elif Content[0] in ['names', 'grouptp']:
            Content[1] = Content[1][1:-1]
            Content[1] = Content[1].replace(' ', '')
            Content[1] = Content[1].replace("'", "")
            Content[1] = Content[1].split(',')
            Content[1][0] = Content[1][0].replace("'", "")
        elif Content[0] == 'intercept' or Content[0] == 'slope':
            if '+/-' in Content[1]:
                Content[1] = Content[1].split('+/-')
                Content[1] = ufloat(float(Content[1][0]), float(Content[1][1]))
            else:
                Content[1] = float(Content[1])
        elif Content[0] == 'XYdata':
            Content[1] = Content[1].split('], [')
            Xdata = Content[1][0].replace('[', '')
            Ydata = Content[1][1].replace(']', '')
            Xdata = Xdata.split(',')
            Ydata = Ydata.split(',')
            
            def TEMP(LIST):
                Ans = copy.deepcopy(LIST)
                i = 0
                for ele in LIST:
                    temp = copy.deepcopy(ele)
                    if '+/-' in temp:
                        temp = temp.split('+/-')
                        Ans[i] = ufloat(float(temp[0]), float(temp[1]))
                    else:
                        Ans[i] = float(temp)
                    i = i + 1
                return Ans
            Content[1] = [TEMP(Xdata), TEMP(Ydata)]
        return Content


    def writeshvartext(self):
        # 텍스트 파일 기본 디렉토리
        if platform.system() == 'Windows':
            BaseDirectory = __file__.replace("exp.py", "var\\")
            if os.path.isdir(BaseDirectory.replace('var\\', 'var')) is False:
                os.mkdir(BaseDirectory.replace('var\\', 'var'))
        else:
            BaseDirectory = __file__.replace('exp.py', 'var/')
            if os.path.isdir(BaseDirectory.replace('var/', 'var')) is False:
                os.mkdir(BaseDirectory.replace('var/', 'var'))
        # Shvar 타입의 variable 정보를 저장할 텍스트파일을 생성

        # 만약 varname이 존재하지 않는다면 쓰지 않는것으로 결정한다.
            # 쓰지 않는 것을 알린다.
        if self.varname is None:
            print(' ')
            print("There is no object name!!")
            print(' ')
            return None
        
        # 오늘의 날짜를 불러온다.
        date, time, Titletime = Shvar.change_time_to_text()
        # date : 날짜, 년-월-일
        # time : 시간, 시:분:초
        # Titletime : 파일 제목용 시간, 시;30분단위분;00

        # 검색하는 파트
            # return True if path is an existing directory
        # 오늘의 날짜에 해당하는 폴더가 있는지 검색한다.
            # 만약에 있다면 그곳에 저장한다.
            # 만약에 없다면 폴더를 생성하고 그곳에 저장한다.
        if os.path.isdir(BaseDirectory + date) is False:
            os.mkdir(BaseDirectory + date)

        # 경로에서 파일 이름들을 불러오고, varname으로 시작하는 파일들을 검색한다.
            # 있다면 있는 것이다.

        # 텍스트 내용을 정한다.
            # 텍스트 내용은 class variable로 한다.
            # 마지막에 작성 시간을 추가한다.
        backdata = [None for i in range(22)]
        backdata[0] = self.tp
        backdata[1] = self.origin
        backdata[2] = self.eachlength
        backdata[3] = self.tbe
        backdata[4] = self.mean
        backdata[5] = self.vrc
        backdata[6] = self.stdv
        backdata[7] = self.SE
        backdata[8] = self.uf
        backdata[9] = self.iftval
        backdata[10] = self.tval
        backdata[11] = self.length
        backdata[12] = self.dsc
        backdata[13] = self.scalar
        backdata[14] = self.scalarlist
        if self.intercept is not None:
            backdata[15] = '{:.16}'.format(self.intercept)
        if self.slope is not None:
            backdata[16] = '{:.16}'.format(self.slope)
        backdata[17] = self.R2
        backdata[18] = self.names
        backdata[19] = self.grouptp
        backdata[20] = self.XYdata
        backdata[21] = self.varname

        content = []
        i = 0
        while i < len(backdata):
            if backdata[i] is not None:
                content.append(Shvar.ClassVarList[i] + '; ' + str(backdata[i]))
            i += 1

        # 텍스트 제목을 정한다.
            # 텍스트의 제목은 varname_시간 으로 결정한다.
            # 시간은 30분 간격이다.
            # 시간은 날짜를 포함한다.
    
        title = str(self.varname) + '_' + Titletime

        # 저장 경로 폴더에 해당 제목의 파일이 있는지 검사한다.
            # 만약에 있다면 덮어쓰기.
            # 만약에 없다면 생성하기.
        if platform.system() == 'Windows':
            f = open(BaseDirectory + date + '\\' + title + '.txt', 'w')
        else:
            f = open(BaseDirectory + date + '/' + title + '.txt', 'w')
        
        time = time.replace(';', ':')
        
        for x in content:
            f.write(x + '\n')
        f.write('Last fixed time : ' + date + ' ' + time)
        f.close()
    

    def readshvartext(varname, Date='today', Time='latest', imp=False):
        

        """
        A method call information of Shvar variable which saved at an backup text file.

        Parameter
            varname : The name of varible you want to call. String.
            Date    : Date when the variable you want to call created.
                      'today', 'yesterday' or 8 digits integer : year(4)month(2)day(2)
            Time    : Time when the variable tou want to call created.
                      Represented in units of 30 minutes.
                      'latest' or 4 digits integer : hour(2)minute(2)
            imp     : Whether import or not. bullen type.
        """

        
        # 텍스트 파일 기본 디렉토리
        if platform.system() == 'Windows':
            BaseDirectory = __file__.replace("exp.py", "var\\")
            if os.path.isdir(BaseDirectory.replace('var\\', 'var')) is False:
                os.mkdir(BaseDirectory.replace('var\\', 'var'))
        else:
            BaseDirectory = __file__.replace('exp.py', 'var/')
            if os.path.isdir(BaseDirectory.replace('var/', 'var')) is False:
                os.mkdir(BaseDirectory.replace('var/', 'var'))

        # 시간을 불러온다.
        date, time, Titletime = Shvar.change_time_to_text()

        # 원하는 날짜를 입력할 수 있게 한다.
            # 기본값은 오늘
            # 형식은 '년월일'
            # 년은 4자리, 월과 일은 2자리
            # 총 8자리
        if Date == 'today':  # 오늘일 경우의 디렉토리 이름
            Directorytitle = date
        elif Date == 'yesterday':  # 어제일 경우의 디렉토리 이름
            Directorytitle = Shvar.change_time_to_text(deltime=1000000, timedeltype='past')[0]
        else:  # 날짜를 직접 입력하는 경우, 날짜는 정수여야한다.
            if isinstance(Date, int):
                if len(str(Date)) <= 8:  # 날짜는 8자리보다 작아야 한다.
                    zerosnum = 8 - len(str(Date))
                    zeros = ''
                    i = 0
                    while i < zerosnum:
                        zeros = zeros + '0'
                        i = i + 1
                    strDate = zeros + str(Date)  # 년도의 입력때문에 8자리가 아닌 입력을 8자리 스트링으로 전환한다.

                    Year = int(strDate[:4])  # 8자리 스트링을 잘라 년도와 월, 일을 의미하는 정수로 바꾼다.
                    Month = int(strDate[4:6])
                    if (Month >= 1 and Month <= 12) is False:  # 월의 입력을 검사한다.
                        raise TypeError("Month should be in 01 ~ 12.")
                    Day = int(strDate[6:])
                    
                    # 윤년을 고려해서 일의 입력이 정상인지 검사한다.
                    if Month == 2:
                        if Shvar.check_leap_year(Year) is True:
                            if (Day >= 1 and Day <= 29) is False:
                                raise ValueError("Day should be in 01 ~ 29.")
                        else:
                            if (Day >= 1 and Day <= 28) is False:
                                raise ValueError("Day should be in 01 ~ 28.")
                    elif Month in [1, 3, 5, 7, 8, 10, 12]:  # 날짜가 31까지만 있음
                        if (Day >= 1 and Day <= 31) is False:
                            raise ValueError("Day should be in 01 ~ 31.")
                    elif Month in [4, 6, 9, 11]:  # 날짜가 30까지만 있음
                        if (Day >= 1 and Day <= 30) is False:
                            raise ValueError("Day should be in 01 ~ 30.")
                    else:
                        raise ValueError("Something is Wrong at Month value.")
                    Year = str(Year)  # 다시 년, 월, 일을 스트링으로 바꿔준다.
                    Month = Shvar.change_one_to_two_digits(str(Month))
                    Day = Shvar.change_one_to_two_digits(str(Day))
                    Directorytitle = Year + '-' + Month + '-' + Day  # 지정날짜
                else:
                    raise TypeError("Date should be 'today', 'yesterday', or 8 or less digits numbers.")
            else:
                raise TypeError("Date should be 'today', 'yesterday', or 8 or less digits numbers.")
        
        if os.path.isdir(BaseDirectory + Directorytitle):  # 지정된 디렉토리가 있는지 본다.
            tempFileList = os.listdir(BaseDirectory + Directorytitle)  # 지정된 디렉토리의 파일들의 리스트를 불러온다.
            FileList = []
            i = 0
            while i < len(tempFileList):
                if tempFileList[i][-4:] == '.txt':
                    FileList.append(tempFileList[i])
                i = i + 1
            varnameList = []  # 앞으로 채워 나갈 리스트이며, FileList에 해당하는 varname리스트다.
            HourList = []  # FileList에 해당하는 Hour리스트다.
            MinuteList = []  # FileList에 해당하는 Minute리스트다.
            i = 0
            while i < len(FileList):  # FileList의 내용들을 varname, Hour, Minute로 찢어서 다시 저장한다.
                FileList[i] = Shvar.split_Filename(FileList[i])
                if FileList[i][0] == varname:
                    varnameList.append(FileList[i][0])
                    HourList.append(FileList[i][1])
                    MinuteList.append(FileList[i][2])
                i = i + 1
            if len(varnameList) == 0:  # 만약 디렉토리에 아무것도 없다면 없다고 알려주고 함수를 종료한다.
                print("There is no file about %s"%(varname))
                return None
        else:
            print("There is no " + "'" + Directorytitle + "'.")  # 해당 디렉토리가 없으면 없다고 알려주고 함수를 종료한다.
            return None
        if Time == 'latest':  # 입력 시간을 '최신'으로 했을 경우
            HourList = [int(x) for x in HourList]  # 해당 디렉토리의 파일이름의 시간대를 정수로 다시 저장한다.
            # 여기까지 온 시점에서 HourList는 비어있을 수 없다.
            MinuteList = [int(x) for x in MinuteList]  # 해당 디렉토리의 파일이름의 분을 정수로 다시 저장한다.
            AvaillatestMinuteList = []  # 채워나갈 리스트이며, 가장 숫자가 큰 시간에 해당하는 분을 저장한다.
            i = 0
            while i < len(HourList):  # 가장 숫자가 큰 시간에 해당하는 분을 저장한다.
                if HourList[i] == max(HourList):
                    AvaillatestMinuteList.append(MinuteList[i])
                i = i + 1
            Hour = Shvar.change_one_to_two_digits(str(max(HourList)))
            TitleMinute = Shvar.change_one_to_two_digits(str(max(AvaillatestMinuteList)))
            Titletime = Hour + ';' + TitleMinute + ';00'
        else:  # 지정된 시간을 입력한 경우
            if isinstance(Time, int) and len(str(Time)) <= 4:  # 입력은 시간분을 연달아 입력한 4자리 정수여야 한다.
                zerosnum = 4 - len(str(Time))  # 입력된 정수는 시간값에 의해 3자리일 수 있으며, 이를 4자리 스트링으로 전환한다.
                zeros = ''
                i = 0
                while i < zerosnum:
                    zeros = zeros + '0'
                    i = i + 1
                strTime = zeros + str(Time)
                Hour = int(strTime[:2])  # 직접 입력한 시간
                TitleMinute = int(int(strTime[2:]) / 30) * 30  # 직접 입력한 분
                if (Hour >= 0 and Hour < 24) is False:  # 시간값 검사
                    raise ValueError("Hour must be in 0 ~ 23")
                if (TitleMinute == 0 or TitleMinute == 30) is False:  # 분값 검사
                    raise ValueError("Minute must be in 0 ~ 59")
                
                Hour = Shvar.change_one_to_two_digits(str(Hour))
                TitleMinute = Shvar.change_one_to_two_digits(str(TitleMinute))
                Titletime = Hour + ';' + TitleMinute + ';00'
                
                AvailableMinuteList = []
                i = 0
                while i < len(HourList):  # 입력한 시간에 해당하는 값이 HourList에 있는지 검사
                    if HourList[i] == Hour:
                        AvailableMinuteList.append(MinuteList[i])
                    i = i + 1
                
                if len(AvailableMinuteList) == 0:  # 입력한 시간에 해당하는 값이 없으면 없다고 알리고 종료
                    print("There is no file about " + varname + '_' + Titletime + '.txt')
                    return None
                else:
                    if (TitleMinute in AvailableMinuteList) is False:
                        # 입력한 분에 해당하는 파일이 있는지 검사, 없으면 없다고 알리고 종료
                        print("There is no file about " + varname + '_' + Titletime + '.txt')
                        return None
            else:
                raise TypeError("Time should be 'latest' or 4 digits numbers.")
        
        Filename = varname + '_' + Titletime + '.txt'
        # 선택된 파일로부터 Shvar object의 값을 불러온다.

        result = Shvar()
        if platform.system == 'Windows':
            f = open(BaseDirectory + Directorytitle + '\\' + Filename, 'r')
        else:
            f = open(BaseDirectory + Directorytitle + '/' + Filename, 'r')
        lines = f.readlines()
        for x in lines:
            x = Shvar.split_Content(x)
            if x[0] == 'tp':
                result.tp = x[1]
            elif x[0] == 'origin':
                result.origin = x[1]
            elif x[0] == 'eachlength':
                result.eachlength = x[1]
            elif x[0] == 'tbe':
                result.tbe = x[1]
            elif x[0] == 'mean':
                result.mean = x[1]
            elif x[0] == 'vrc':
                result.vrc = x[1]
            elif x[0] == 'stdv':
                result.stdv = x[1]
            elif x[0] == 'SE':
                result.SE = x[1]
            elif x[0] == 'uf':
                result.uf = x[1]
            elif x[0] == 'iftval':
                result.iftval = x[1]
            elif x[0] == 'tval':
                result.tval = x[1]
            elif x[0] == 'length':
                result.length = x[1]
            elif x[0] == 'dsc':
                result.dsc = x[1]
            elif x[0] == 'scalar':
                result.scalar = x[1]
            elif x[0] == 'scalarlist':
                result.scalarlist = x[1]
            elif x[0] == 'intercept':
                result.intercept = x[1]
            elif x[0] == 'slope':
                result.slope = x[1]
            elif x[0] == 'R2':
                result.R2 = x[1]
            elif x[0] == 'names':
                result.names = x[1]
            elif x[0] == 'grouptp':
                result.grouptp = x[1]
            elif x[0] == 'XYdata':
                result.XYdata = x[1]
            elif x[0] == 'varname':
                result.varname = x[1]
        
        print("\n[Read Result]")
        result.printfull()
        print("")
        
        if imp is True:
            # 텍스트의 내용을 앞부분과 뒷부분으로 찢은 뒤,
            # 앞의 내용에 맞춰서 result라는 새로운 shvar object에 삽입
            # 최종적으로 varname에 맞춰서 result에 동기화
            callerframerecord = inspect.stack()[1]
            frame = callerframerecord[0]
            info = inspect.getframeinfo(frame)
            context = info.code_context[0]  # 이 코드가 실행된 프레임에서 해당 줄을 뽑아온다.
            context = context.replace('\n', '')  # 엔터와 띄어쓰기를 전부 없앤다.
            if '=' in context:  # 만약 '='을 이용해 readshvartext의 return을 어느 object에 부여했다면
                eqnum = context.count('=')
                if eqnum == 1:  # 보통은 '='는 한 개일 것이며,
                    context = context.replace(' ', '')
                    tempvarname = context.split('=')[0]  # varname을 뽑아낸다.
                else:  # 만약 '='가 두 개 이상이면, 세미콜론을 이용한 코드라고 가정
                    if context.count('readshvartext') == 1:  # readshvartext를 한 번만 사용한 경우면 봐준다.
                        context = context.split('readshvartext')[0]  # readshvartext를 찾아서 앞부분만 추출
                        context = context.split(';')[-1]  # 세미콜론이 있다면 뒷부분을 추출
                        context = context.split('=')[0]  # '='을 찾아서 앞부분을 추출
                        tempvarname = context.replace(' ', '')  # 띄어쓰기까지 없애서 저장
                    else:  # readshvartext를 두 번 쓴 것이라면 가차없이 용서하지 않는다.
                        raise TypeError("Why are you using semicolon as calling this method, 'readshvartext'!")
            result.varname = tempvarname  # 이 부분은 백업 안되어있음 다음 버젼에 올리는 걸로.
            result.writeshvartext()
        return result



import traceback, threading, time

class InstanceCreationError(Exception):
    pass

class RememberInstanceCreationInfo:
    def __init__(self):
        for frame, line in traceback.walk_stack(None):
            varnames = frame.f_code.co_varnames
            if varnames is ():
                break
            if frame.f_locals[varnames[0]] not in (self, self.__class__):
                break
                # if the frame is inside a method of this instance,
                # the first argument usually contains either the instance or
                #  its class
                # we want to find the first frame, where this is not the case
        else:
            pass
            # raise InstanceCreationError("No suitable outer frame found.")
        self._outer_frame = frame
        self.creation_module = frame.f_globals["__name__"]
        self.creation_file, self.creation_line, self.creation_function, \
            self.creation_text = \
            traceback.extract_stack(frame, 1)[0]
        self.creation_name = self.creation_text.split("=")[0].strip()
        super().__init__()
        # threading.Thread(target=self._check_existence_after_creation).start()
"""
    def _check_existence_after_creation(self):
        while self._outer_frame.f_lineno == self.creation_line:
            time.sleep(0.01)
        # this is executed as soon as the line number changes
        # now we can be sure the instance was actually created
        error = InstanceCreationError(
                "\nCreation name not found in creation frame.\ncreation_file: "
                "%s \ncreation_line: %s \ncreation_text: %s\ncreation_name ("
                "might be wrong): %s" % (
                    self.creation_file, self.creation_line, self.creation_text,
                    self.creation_name))
        nameparts = self.creation_name.split(".")
        try:
            var = self._outer_frame.f_locals[nameparts[0]]
        except KeyError:
            raise error
        finally:
            del self._outer_frame
        # make sure we have no permament inter frame reference
        # which could hinder garbage collection
        try:
            for name in nameparts[1:]: var = getattr(var, name)
        except AttributeError:
            raise error
        if var is not self: raise error

    def __repr__(self):
        return super().__repr__()[
               :-1] + " with creation_name '%s'>" % self.creation_name
"""